/*=========================================================================

  Program:   Orbit Visualization Tool
  Source:    $Source: /net/nest/stor/devel/ovt2g/ovt/mag/MagActivityDataModel.java,v $
  Date:      $Date: 2000/04/12 13:03:25 $
  Version:   $Revision: 1.8 $


Copyright (c) 2000 OVT Team (Kristof Stasiewicz, Mykola Khotyaintsev, 
Yuri Khotyaintsev)
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification is permitted provided that the following conditions are met:

 * No part of the software can be included in any commercial package without
written consent from the OVT team.

 * Redistributions of the source or binary code must retain the above
copyright notice, this list of conditions and the following disclaimer.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS
IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
THE IMPLIED WARRANTIES OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT OR
INDIRECT DAMAGES  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE.

OVT Team (http://ovt.irfu.se)   K. Stasiewicz, M. Khotyaintsev, Y.
Khotyaintsev

=========================================================================*/

/*
 * MagActivityDataModel.java
 *
 * Created on den 25 mars 2000, 00:30
 */

package ovt.mag;

import java.io.*;
import java.util.*;

/** Class for handling magnetic activity data stored in the disk files.
 *Files should exist in directory <I>mdata/</I> on the disk and have
 *header with column names. First column should be always <B>Time</B>.
 *All extra and unsuficcient data will be ignored and overwritten
 *after <CODE>save()</CODE>
 * @author Yuri Khotyaintsev
 * @version 1.0
 */
public class MagActivityDataModel extends javax.swing.table.AbstractTableModel {

  private String name = null;
  private Vector data = new Vector();
  private double minValue;
  private double maxValue;
  private String[] columnNames = null;
  private int columnNumber = 0;
  private int rowNumber = 0;
  private double lastMjd = 0;
  /** Extension for activity file name. Should be <B>.dat</B>
   */
  public static final String magDataExt =".dat";
  /** Extension for activity backup file name. Should be <B>.bak</B>
   */
  public static final String magDataBak =".bak";
  private File file = null;

  /** Creates new MagActivityDataModel
   * @param param Name of activity parameter.
   * @throws Exception for parsing problems
   * @throws FileNotFoundException for file lookup
   * @throws IOException for IO problems
   */
  public MagActivityDataModel(String param, double minValue, double maxValue)
  throws Exception, IOException {
    name = param;
    this.minValue = minValue;
    this.maxValue = maxValue;
    file = new File(ovt.OVTCore.getUserdataDir(), param + magDataExt);
    RandomAccessFile fileIn = new RandomAccessFile(file,"r");
    long length = fileIn.length();
    if ( length <= 0 ) {
      throw new Exception("MagActivityDataModel: activity file is empty");
    }
    // Column Names
    String s = fileIn.readLine();
    if (s.startsWith("Time")){
      StringTokenizer tok = new StringTokenizer(s);
      columnNumber = tok.countTokens();
      //System.out.println("columnNumber: "+columnNumber);
      if (columnNumber < 1 ) {
        throw new Exception("MagActivityDataModel: invalid activity file header");
      }
      columnNames = new String[columnNumber];
      columnNames[1]= "ttt";
      for (int i=0; i < columnNumber; i++) {
        String ss = tok.nextToken();
        //System.out.println("col: "+ i + " value: "+ss);
        columnNames[i] = ss;
      }
    }else{
      throw new Exception("MagActivityDataModel: invalid activity file header");
    }

    /* Read data
    first entry is time in forma 1994-04-04 12:00
    to be understood by ovt.datatype.Time.
    Incomplete lines or extra data will be disregarded*/
    long filePointer = fileIn.getFilePointer();
    int rowCount = 0;
    int fileCount = 1;
    while(filePointer < length) {
      boolean parsed = true;
      fileCount++;
      s = fileIn.readLine();
      StringTokenizer tok = new StringTokenizer(s);
      int cn = tok.countTokens();
      if (cn >= columnNumber + 1){
        Vector line = new Vector(columnNumber);
        String tmpStr = null;
        for (int i=0; i<=columnNumber; i++) {
          String ss = tok.nextToken();
          if (i==0){
            tmpStr = ss;
          } else if (i==1){
            //Time
            //System.out.println("Found: " + tmpStr + " " + ss);
            try {
              ovt.datatype.Time time = new ovt.datatype.Time(tmpStr + " " + ss);
              line.addElement(new Double(time.getMjd()));
            } catch (IllegalArgumentException e) {
              System.out.println("parse error in line #" + fileCount + " file "
              + file.getAbsolutePath());
              parsed = false;
              break;
            }
          } else {
            //Any digital data
            try {
              Double dd = new Double(ss);
              if ( isValid(dd) ) line.addElement(dd);
              else {
                System.out.println("value out of range in line #" + fileCount
                + " file " + file.getAbsolutePath());
                parsed = false;
                break;
              }
            } catch (NumberFormatException e) {
              System.out.println("parse error in line #" + fileCount + " file "
              + file.getAbsolutePath());
              parsed = false;
              break;
            }
          }
        }
        if (parsed) {
          data.addElement(line);
          rowCount++;
        }
      }
      filePointer = fileIn.getFilePointer();
    }
    rowNumber = data.size();
    fileIn.close();

    sortData();
    lastMjd = getMjd(0);
  }

  /**
   * @return number of rows
   */
  public int getRowCount() {
    return rowNumber;
  }

  public int getColumnCount(){
    return columnNumber;
  }

  public Object getValueAt(int row, int col){
    if( row<0 || row>=rowNumber || col<0 || col>=columnNumber)
    throw new IllegalArgumentException("Index out of bounds");
    else {
      if(col == 0) {
        Vector vec = (Vector)data.elementAt(row);
        Double mjd = (Double)vec.elementAt(0);
        return ovt.datatype.Time.toString(mjd.doubleValue());
      } else {
        Vector vec = (Vector)data.elementAt(row);
        return vec.elementAt(col);
      }
    }
  }

  public Object getValueAt(double mjd,int element) {
    if ( mjd > getMjd(0) ) {
      int i = 0;
      while ( mjd > getMjd(i) ) {
        if ( i == rowNumber-1 ) break;
        i++;
      }
      lastMjd = getMjd(i);
      return getValueAt(i,element);
    } else return null;
  }

  public double getMjd(int row) {
    Vector vec = (Vector)data.elementAt(row);
    Double mjd = (Double)vec.elementAt(0);
    return mjd.doubleValue();
  }

  public double getLastMjd() {
    return lastMjd;
  }

  /** returns data
   * @param row row
   * @param col column
   * @throws InvalidParameterException
   * @return
   */
  public double getDataAt(int row,int col) throws java.security.InvalidParameterException {
    if( row<0 || row>=rowNumber || col<0 || col>=columnNumber)
    throw new java.security.InvalidParameterException("index out of range");
    else {
      Vector vec = (Vector)data.elementAt(row);
      Double dd = (Double)vec.elementAt(col);
      return dd.doubleValue();
    }
  }

  public boolean isCellEditable(int row, int col){
    return true;
  }

  public void setValueAt(Object value, int row, int col) {
    if(col == 0) {
      if (ovt.datatype.Time.isValid((String)value)){
      Vector vec = (Vector)data.elementAt(row);
      Double mjd = new Double( ovt.datatype.Time.getMjd((String)value));
      vec.setElementAt(mjd,0);
      if ( row < rowNumber-1 )
      if ( mjd.doubleValue() > getMjd(row+1) ){
        int i = row+1;
        while ( mjd.doubleValue() > getMjd(i) ) {
          flipRows(i,i-1);
          fireTableRowsUpdated(i-1,i);
          i++;
          if (i == rowNumber) break;
        }
      } else if ( row > 0 )
      if ( mjd.doubleValue() < getMjd(row-1) ){
        int i = row-1;
        while ( mjd.doubleValue() < getMjd(i) && i >= 0) {
          flipRows(i,i+1);
          fireTableRowsUpdated(i,i+1);
          i--;
          if ( i < 0 ) break;
        }
      }
    }
    } else {
      Vector vec = (Vector)data.elementAt(row);
      try {
        Double dat = new Double((String)value);
        if ( isValid(dat) ) vec.setElementAt(dat,col);
        else System.out.println("Specified value is out of range");
      } catch (NumberFormatException e) {
        System.out.println("Invalid number format");
      }
    }
  }

  public String getName(){
    return name;
  }

  public String getColumnName(int col) {
    if (col<0 || col>=columnNumber) return null;
    else return columnNames[col].toString();
  }

  /** Saves <CODE>data</CODE> to file. Old file should be moved to .bak
   * @throws FileNotFoundException {@link java.io.FileNotFoundException}
   * @throws IOException {@link java.io.IOException}
   */
  public void save() throws FileNotFoundException, IOException {
    //System.out.println("Saving ...");
    // backup file
    if (file.exists()) {
      File bak = new File(ovt.OVTCore.getUserdataDir(), name + magDataBak);
      file.renameTo(bak);
      file = new File(ovt.OVTCore.getUserdataDir(), name + magDataExt);
    }
    RandomAccessFile fileOut = new RandomAccessFile(file,"rw");
    // create header
    String line = "Time\t\t";
    for (int i=1; i<columnNumber; i++) {
      line = line + "\t"+ columnNames[i];
    }
    fileOut.writeBytes(line + "\n");
    // write data
    for (int i=0; i<rowNumber; i++) {
      line = (String)getValueAt(i,0);
      for (int j=1; j<columnNumber; j++) {
        Double dd = (Double)getValueAt(i,j);
        line = line + "\t"+ dd.toString();
      }
      //line = line + "\n";
      fileOut.writeBytes(line + "\n");
    }
    fileOut.close();
    //System.out.println("Done");
  }

  public void insertRows(int row){
    Vector vec = (Vector)data.elementAt(row);
    Vector nv = (Vector)vec.clone();
    data.insertElementAt(nv,row+1);
    rowNumber = data.size();
    fireTableRowsInserted(row, row + 1);
  }

  public void deleteRows(int firstRow, int lastRow){
    for(int i=lastRow;i>=firstRow;i--){
      data.removeElementAt(i);
    }
    rowNumber = data.size();
    fireTableRowsDeleted(firstRow, lastRow);
  }


  protected void flipRows(int a, int b){
    Vector veca = (Vector)data.elementAt(a);
    Vector vecb = (Vector)data.elementAt(b);
    data.setElementAt(veca,b);
    data.setElementAt(vecb,a);
  }

  protected void sortData() {
    int lo = 0;
    int up = rowNumber-1;
    int i,j;
    while ( up > lo ) {
      j = lo;
      for ( i = lo; i < up; i++ ){
        if( getMjd(i) > getMjd(i+1) ) {
          flipRows(i,i+1);
          j = i;
        }
      }
      up = j;
      for ( i = up; i > lo; i-- ){
        if( getMjd(i) < getMjd(i-1) ) {
          flipRows(i,i-1);
          j = i;
        }
      }
      lo = j;
    }

  }

  public boolean isValid(double value) {
    if ( value >= minValue && value <= maxValue ) return true;
    else return false;
  }

  public boolean isValid(Double value) {
    double val = value.doubleValue();
    if ( val >= minValue && val <= maxValue ) return true;
    else return false;
  }
}
