
/*----------------------------------------------------------------
This collection of routines must stay together in the same source
file because of the static declarations below. The intention is to
hide all the garbage parameters from the caller.

The original FORTRAN verision assumed that all the local variables
survived between calls, which doesn't work if they are on a stack as
in a C environment. Consequently, I've put all the locals from the
original SGP4 into a private static array.
----------------------------------------------------------------*/
#include "extern.h"
#include <ctype.h>
#include <math.h>
#include <jni.h>
#include "ovt_object_Tle.h"
#include "globvar.h"

struct satco_ satco[2];
static char line1[100];          /* for reading element file */
static char line2[100];          /* for reading element file */
static char line[100];

static double   ae = 1.0;
static double   de2ra = PI / 180.;
static double   qo = 120.0;
static double   so = 78.0;
static double   tothrd = 2. / 3.;
static double   twopi = PI *2.;
static double   xj2 = 1.082616e-3;
static double   xj3 = -.253881e-5;
static double   xj4 = -1.65597e-6;
static double   xke = 0.743669161e-1;
static double   xkmper = R_EARTH;
static double   Mnpda = MIN_PER_DAY;

/*----------------------------------------------------------------
The sgp4 routine was plagarized from the FORTRAN code in the
spacetrack #3 report

The following routines are used to read the orbital elements from
a file in the standard NASA format that looks like:

name of satellite
1 88888U          80275.98708465  .00073094  13844-3  66818-4 0    8
2 88888  72.8435 115.9689 0086731  52.6988 110.5714 16.05824518  105

----------------------------------------------------------------*/
double  get_d (i, j)            /* extract double from "line" */
int i,                        /* start column #, 1-rel */
j;                        /* end column #, 1-rel */
{
    int k;
    double  x;
    line[j] = 0;                  /* terminate */
    x = 0.;                       /* in case field is blank */
    k = sscanf (&line[i - 1], "%lf", &x);
    return (x);
}


long    get_l (i, j)              /* extract long from "line" */
int i,                        /* start column #, 1-rel */
j;                        /* end column #, 1-rel */
{
    int k;
    long    x;
    line[j] = 0;                  /* terminate */
    x = 0;                        /* in case field is blank */
    k = sscanf (&line[i - 1], "%ld", &x);
    return (x);
}


blank (line)                    /* init line with blanks */
char    *line;
{                   /* short lines will leave trash */
    int i;
    for (i = 0; i < 100; i++, *line++ = ' ')
        ;
}


check_sum (s)
char    s[71];
{
    int i, chksum;
    char    c;

    for (i = 0, chksum = 0; i <= 67; i++) {
        c = s[i];
        if (isdigit (c))            /* 0-9? */
            chksum += c - '0';
        else if (c == '-')
            chksum += 1;
    }
    chksum = chksum % 10;         /* make modulo 10 */
    c = s[68] - '0';              /* typed in checksum */
    if (chksum != c) {            /* error? */
/*    // *****Commented becouse of "stdout BUG"
        printf ("\n%c%s\ncalc chksum=%d, typed=%d\n", 7, s, chksum, c);

*/        return (1);                 /* error */
    }
    return (0);                   /* OK */
}


NgetElements (infile, jul_epoch, relem, end_flag)
FILE *infile;
double  relem[9];
struct julian *jul_epoch;
int *end_flag;

{
    int i;
    double  day, iexp, ibexp, xxnodeo;
    double  xxndt2o, xxndd6o, xbstar, xxincl, xeo, xomegao, xxmo, xxno;
    long    obj,year;
    /*----------------------------------------------------------------
   Read in orbital elements from file. The get_d and get_l routines
   should read the fields from line BACKWARDS, since they insert 
   terminating nulls at the end of each field extracted.
----------------------------------------------------------------*/

    *end_flag = 0;
    blank (line);
    if (fgets (line, 100, infile) == NULL) {
        *end_flag = 1;
        return;
    }

    if ((debug_flag > 2))
        printf (" line #1 %s\n", line);

    check_sum (line);             /* verify checksum */

    i = 0;
    sscanf (line, "%1d", &i);
    strcpy(line1,line);
    if (i != 1) {                 /* this should be line # 1 */
        printf (" line #1 %s\n", line);
        printf ("phase error, i=%d, should be 1\n", i);
        exit (1);
    }
    xbstar = 0.;
    xxndd6o = 0.;

    ibexp = get_d (60, 61);
    xbstar = 0.00001 * get_d (54, 59);
    xbstar = xbstar * pow (10., ibexp) / ae;
    iexp = get_d (51, 52);
    xxndd6o = 0.00001 * get_d (45, 50);
    xxndd6o = xxndd6o * (pow (10.0, iexp));

    xxndt2o = get_d (34, 43);
    day = get_d (21, 32);
    obj = get_l (3, 7);
    year = get_l (19, 20);
    if(year>50)
       year += 1900;
    else year+=2000;
    julian_day (year, 1, 1, 0., 0., 0., jul_epoch);
    jul_epoch->integer += (int)(day - 1.0);
    jul_epoch->fraction = fmod(day, 1.0e0);

    if ((debug_flag)) {
        printf ("%s\n", sat_name);
        printf ("obj=%ld year=%ld day=%lg xndt2o=%lg xndd6o=%lg\n",
            obj, year, day, xxndt2o, xxndd6o);
        printf(" jul_ep = %lf  %lf\n", jul_epoch->integer, jul_epoch->fraction);
    }

    /* line 2 */

    blank (line);
    if (fgets (line, 100, infile) == NULL) {
        printf (" end of element file encountered\n");
        *end_flag = 1;
        return;
    }
    strcpy(line2,line);
    if ((debug_flag > 2))
        printf (" line #2 %s\n", line);

    check_sum (line);             /* verify checksum */

    sscanf (line, "%1d", &i);
    if (i != 2) {
        printf ("phase error, i=%d, should be 2\n", i);
        printf (" line #2 %s\n", line);
        exit (1);
    }
    xxno = get_d (53, 63);
    xxmo = get_d (44, 51);
    xomegao = get_d (35, 42);
    xeo = get_d (27, 33);
    xeo *= .0000001;
    xxnodeo = get_d (18, 25);
    xxincl = get_d (9, 16);
    relem[0] = xxndt2o;
    relem[1] = xxndd6o;
    relem[2] = xbstar;
    relem[3] = xxincl;
    relem[4] = xxnodeo;
    relem[5] = xeo;
    relem[6] = xomegao;
    relem[7] = xxmo;
    relem[8] = xxno;
}

int Nread_elements(filename, mjd, mjd1, mjd2)
char    *filename;
double  mjd;     /* time for requested attitude */
double  *mjd1;   /* start of the epoch elements  */
double  *mjd2;   /* next elements epoch */
{
    FILE * file;
    int i, ymd, hms;
    static int  ms = 0;
    struct julian jepoch;
    static int  endFlag = 0;
    static double   relem1[9], relem2[9];
    double day,gap;
    int ind,tmp_i=0;
    int NgetElements();
      static int first=0;
      static double mjdlast;
      FILE *fo;
      mjdlast=0.0;

    if (mjd < *mjd1 || mjd > *mjd2) {
        if ((file = fopen(filename, "r")) == NULL) {
            printf("cannot open orbit elements file %s\n", filename);
            exit(-1);
        }
        *mjd1 = -1.0;
        *mjd2 = -1.0;

   /*     if (satid == 0)
            ind = 0;
        else */
            ind = 1;

        fgets(line, 79, file);

        sscanf(line, "%s", sat_name);
        endFlag = 0;

        NgetElements(file, &jepoch, relem1, &endFlag);
        mjd2jul(mjd1, &jepoch, -1);
        if (debug_flag > 2) {
            printf("Read first element set\n");
            printf("orbit request for mjd=%lf  first mjd1=%lf\n", mjd, *mjd1);
        }
        if (mjd>10 && mjd < *mjd1 || endFlag != 0) {
            printf("orbit request for time earlier than the start of data\n");
            *mjd2 = *mjd1;
            return (-1);
        }

        while (mjd > *mjd2) {
            NgetElements(file, &jepoch, relem2, &endFlag);
            if (endFlag != 0) {
                *mjd2 = 1e9;
                /*printf("WARNING: orbit extrapolated %.1lf days\n", (mjd - *mjd1));*/
                cdate(&ymd, &hms, &ms, mjd1, -1);
                /*printf("WARNING: last orbital data:    %d %06d\n", ymd, hms);*/
            } else {
                mjd2jul(mjd2, &jepoch, -1);
                if (mjd > *mjd2) {
		    gap= *mjd2 - *mjd1;
		    if(gap<0.5|| gap>14.0){ 
		    mjd2day(mjd1,&day,1);
		    /*
		    if(gap>14.0)
		    printf("orbital data gap %.1lf file %s at %.6lf \n",
		    gap,filename,day);
		    */
/*		    if(gap<0.1)  // *****Commented becouse of "stdout BUG"
		    printf("incorrect sequence in file %s at %.6lf\n",
		    filename,day);
                    fflush(stdout);*/
		     }
                   /* Commented by KS
                     if(first==0) fo=fopen("out","w");
                     first=1;
		     if((*mjd2-mjdlast)>4.0) {
		     fprintf(fo,"%s",line1);
		     fprintf(fo,"%s",line2);
		     mjdlast= *mjd2;
		     }
		     */
                    *mjd1 = *mjd2;
                    for (i = 0; i < 9; i++)
                        relem1[i] = relem2[i];
                }
                if (debug_flag > 1) {
                    cdate(&ymd, &hms, &ms, mjd1, -1);
                    print_jtime(stdout, &jepoch);
                    printf(" %d %d mjd1=%lf  mjd2= %lf\n", ymd, hms, *mjd1, *mjd2);
                    printf(" jepoch= %lf %lf\n", jepoch.integer, jepoch.fraction);
                }
            }
        }   /* while */
        fclose(file);
        NinitiateSgp(relem1, &satco[ind]);
	return 1;
    }
}


/*---------- JNI for NgetSatPos -------------------*/

JNIEXPORT jint JNICALL 
Java_ovt_object_Tle_getSatPosJNI(env, obj, jfilename, jmjd, jgei, jvei)
JNIEnv *env;
jobject obj;
jstring jfilename;
jdouble jmjd;
jdoubleArray jgei;
jdoubleArray jvei;

{

  int i;
  jint result = 0;

  jdouble *gei = (*env)->GetDoubleArrayElements(env, jgei, 0);
  jdouble *vei = (*env)->GetDoubleArrayElements(env, jvei, 0);
  const char *filename = (*env)->GetStringUTFChars(env, jfilename, 0);
  
  
/*  printf("\nmjd = %f\n", jmjd);  */

  result = NgetSatPos(filename, jmjd, gei, vei);

 /* for (i=0; i<3; i++)    printf("c: gei[%d] = %f\n",i,gei[i]); */
  
  (*env)->ReleaseStringUTFChars(env, jfilename, filename);  
  (*env)->ReleaseDoubleArrayElements(env, jgei, gei, 0);
  (*env)->ReleaseDoubleArrayElements(env, jvei, vei, 0);

  return result;
}

int NgetSatPos(filename, mjd, gei, vei)
double  mjd;
double  gei[3], vei[3];
char    *filename;
{
    static double   mjd1[2] = { 0.0, 0.0};
    static double   mjd2[2] = { 0.0, 0.0};
    static int satlast=100;


    double  tsince;
    int i, ind, ret;
    struct elem_ nelem;
    struct julian jday;
    double  sunv[3], gm;
    double  gha(), gmstime();

    debug_flag = 0;
/*    if (satid == 0)
        ind = 0;
    else */
    
        ind = 1;
  /* ind 0 and ind 1 are used for simultaneous conjunction calculations
   normally ind=1 is used */

   /* if(satid*satlast>0 && satid !=satlast) {*/
    mjd1[0]=mjd1[1]=0.0;
    mjd2[0]=mjd2[1]=0.0;
    /*}*/

    if (mjd < mjd1[ind] || mjd > mjd2[ind]){
        ret=Nread_elements(filename, mjd, &mjd1[ind], &mjd2[ind]);
	gm=mjd2[ind]-mjd1[ind];
	if(gm>14.0 && gm<1000.0) printf("orbit element gap %.1lf days\n",gm);
    if(ret<0) {
    for (i = 0; i < 3; i++)
    {
        gei[i] = 1.0;
        vei[i] = 1.0;
    }
    return (-1);
    }
    }
  /*  satlast =satid;*/


    tsince = (mjd -mjd1[ind])
        *1440;  /* minutes from the epoch elements */
    Nsgp4(tsince, &nelem, &satco[ind]);

    for (i = 0; i < 3; i++)
    {
        gei[i] = nelem.position[i];
        vei[i] = nelem.velocity[i];
    }


  /*  if (debug_flag > 1)  This part contains bug.
    {
        sunmjd(mjd, sunv);
        gm = gmstime(mjd) * irad / 15.0;
        mjd2jul(&mjd, &jday, 1);
        gm = gha(jday) / 15.0;
        sun_vect(jday, sunv);
        normf(sunv, 1.00);
        printf(" sunjul= %lf  %lf %lf %lf\n", gm, sunv[0], sunv[1], sunv[2]);
    }*/

    return(0);
}


NinitiateSgp(relem, yy)
double  relem[9];
struct satco_ *yy;
{
    double  a1,  theta2,  eosq, betao2, betao, del1, ao,
    delo,  s4, qoms24, perige, pinvsq, tsi, 
    etasq, eeta, psisq, coef, coef1,   a3ovk2, 
    temp1, temp2, temp3,  x1m5th,
    xhdot1,   
    c1sq,  temp, 
    xno,
    ck4, ae2, ae4, qoms2t, c2,
    c3, theta4 ;

    double  ss;
    int i;

    /*----------------------------------------------------------------
      recover original mean motion (xnodp) and semimajor axis (aodp)
      from input elements
----------------------------------------------------------------*/
    ae2 = ae * ae;
    ae4 = ae2 * ae2;
    yy->ck2 = 0.5 * xj2 * ae2;
    ck4 = -.375 * xj4 * ae4;
    qoms2t = pow (fabs (((qo - so) * ae / xkmper)), (double) 4.);
    ss = ae * (1. + so / xkmper);
    temp = twopi / Mnpda / Mnpda;


    yy->xndt2o = relem[0] * temp;
    yy->xndd6o = relem[1] * temp / Mnpda;
    yy->bstar = relem[2];
    yy->xincl = relem[3] * de2ra;
    yy->xnodeo = relem[4] * de2ra;
    yy->eo = relem[5];
    yy->omegao = relem[6] * de2ra;
    yy->Mo = relem[7] * de2ra;
    xno = relem[8] * temp * Mnpda;
    if (debug_flag > 2) {
        printf(" initializing Sgp...\n");
        for (i = 0; i < 9; i++)
            printf(" relem(%d)=%.9lf\n", i, relem[i]);
    }
    /*
      read_elem ();
*/
    a1 = pow ((xke / xno), tothrd);
    yy->cosio = cos (yy->xincl);
    theta2 = yy->cosio * yy->cosio;
    yy->x3thm1 = 3. * theta2 - 1.;
    eosq = yy->eo * yy->eo;
    betao2 = 1. - eosq;
    betao = sqrt (betao2);
    del1 = 1.5 * yy->ck2 * yy->x3thm1 / (a1 * a1 * betao * betao2);
    ao = a1 * (1. - del1 * (.5 * tothrd + del1 * (1. + 134.
        / 81. * del1)));
    delo = 1.5 * yy->ck2 * yy->x3thm1 / (ao * ao * betao * betao2);
    yy->xnodp = xno / (1. + delo);
    yy->aodp = ao / (1. - delo);

    /*----------------------------------------------------------------
      for perigee less than 220 kilometers, the isimp flag is set and
      the equations are truncated to linear variation in sqrt a and
      quadratic variation in mean anomaly.  also, the c3 term, the
      delta omega term, and the delta m term are dropped.
----------------------------------------------------------------*/

    yy->isimp = 0;
    if ((yy->aodp * (1. - yy->eo) / ae) < (220. / xkmper + ae))
        yy->isimp = 1;
    /*----------------------------------------------------------------
      for perigee below 156 km, the values of
      s and qoms2t are altered
----------------------------------------------------------------*/

    s4 = ss;
    qoms24 = qoms2t;
    perige = (yy->aodp * (1. - yy->eo) - ae) * xkmper;

    if (perige <= 156.) {
        s4 = perige - 78.;
        if (perige <= 98.)
            s4 = 20.;
        qoms24 = pow (((120. - s4) * ae / xkmper), (double) 4);
        s4 = s4 / xkmper + ae;
    }
    pinvsq = 1. / (yy->aodp * yy->aodp * betao2 * betao2);

    tsi = 1. / (yy->aodp - s4);
    yy->eta = yy->aodp * yy->eo * tsi;
    etasq = yy->eta * yy->eta;
    eeta = yy->eo * yy->eta;
    psisq = fabs (1. - etasq);
    coef = qoms24 * pow (tsi, (double) 4.);
    coef1 = coef / pow (psisq, (double) 3.5);
    c2 = coef1 * yy->xnodp * (yy->aodp * (1. + 1.5 * etasq + 
        eeta * (4. + etasq)) + .75 * 
        yy->ck2 * tsi / psisq * yy->x3thm1 * (8. + 3. * etasq * (8. + etasq)));

    yy->c1 = yy->bstar * c2;
    yy->sinio = sin (yy->xincl);
    a3ovk2 = -(xj3 / yy->ck2) * pow (ae, (double) 3.);
    c3 = coef * tsi * a3ovk2 * yy->xnodp * ae * yy->sinio / yy->eo;
    yy->x1mth2 = 1. - theta2;

    yy->c4 = 2. * yy->xnodp * coef1 * yy->aodp * betao2 * (yy->eta * 
        (2. + .5 * etasq) + yy->eo * (.5 + 2. * etasq) - 2. * yy->ck2 * 
        tsi / 
        (yy->aodp * psisq) * (-3. * yy->x3thm1 * (1. - 2. * eeta + etasq * 
        (1.5 - .5 * eeta)) + .75 * yy->x1mth2 * (2. * etasq - eeta * 
        (1. + etasq)) * cos (2. * yy->omegao)));


    yy->c5 = 2. * coef1 * yy->aodp * betao2 * (1. + 2.75 * 
        (etasq + eeta) + eeta * etasq);

    theta4 = theta2 * theta2;
    temp1 = 3. * yy->ck2 * pinvsq * yy->xnodp;
    temp2 = temp1 * yy->ck2 * pinvsq;
    temp3 = 1.25 * ck4 * pinvsq * pinvsq * yy->xnodp;
    yy->Mdot = yy->xnodp + .5 * temp1 * betao * yy->x3thm1 + .0625 * temp2 * 
        betao * 
        (13. - 78. * theta2 + 137. * theta4);
    x1m5th = 1. - 5. * theta2;
    yy->omgdot = -.5 * temp1 * x1m5th + .0625 * temp2 * (7. - 114.
        *theta2 + 
        395. * theta4) + temp3 * (3. - 36. * theta2 + 49. * theta4);
    xhdot1 = -temp1 * yy->cosio;
    yy->xnodot = xhdot1 + (.5 * temp2 * (4. - 19. * theta2) + 2.
        *temp3 * (3. - 
        7. * theta2)) * yy->cosio;
    yy->omgcof = yy->bstar * c3 * cos (yy->omegao);
    yy->Mcof = -tothrd * coef * yy->bstar * ae / eeta;
    yy->xnodcf = 3.5 * betao2 * xhdot1 * yy->c1;
    yy->t2cof = 1.5 * yy->c1;
    yy->Lcof = 0.125 * a3ovk2 * yy->sinio * (3. + 5. * yy->cosio) / (1.
        + yy->cosio);
    yy->aycof = 0.25 * a3ovk2 * yy->sinio;
    yy->delMo = pow ((1. + yy->eta * cos (yy->Mo)), (double) 3.);
    yy->sinMo = sin (yy->Mo);
    yy->x7thm1 = 7. * theta2 - 1.;

    if (!yy->isimp) {
        c1sq = yy->c1 * yy->c1;
        yy->d2 = 4. * yy->aodp * tsi * c1sq;
        temp = yy->d2 * tsi * yy->c1 / 3.;
        yy->d3 = (17. * yy->aodp + s4) * temp;
        yy->d4 = 0.5 * temp * yy->aodp * tsi * (221. * yy->aodp + 31.
            *s4) * yy->c1;
        yy->t3cof = yy->d2 + 2. * c1sq;
        yy->t4cof = 0.25 * (3. * yy->d3 + yy->c1 * (12. * yy->d2 + 10. * 
            c1sq));
        yy->t5cof = 0.2 * (3. * yy->d4 + 12. * yy->c1 * yy->d3 + 6. * yy->d2 * 
            yy->d2 + 
            15. * c1sq * (2. * yy->d2 + c1sq));
    }
}


Nsgp4 (tsince, elem, yy)
double  tsince;                 /* input: time since epoch in minutes */
struct elem_ *elem;             /* output: struct of orbital elemements */
struct satco_ *yy;
{

    double  temp1, temp2, temp3,   
    temp  ;
    double  Mdf, omgadf, xnoddf, omega, Mp, tsq, xnode, tempa, tempe,
    templ, delomg, delm, tcube, tfour, a, e, L, beta, xn, axn, Ll, aynl,
    Lt, ayn, capu, sinepw, cosepw, temp4, temp5, temp6, epw, ecosE,
    esinE, elsq, pl, r, rdot, rfdot, betal, cosu, sinu, u, sin2u, cos2u,
    rk, uk, xnodek, xinck, rdotk, rfdotk, sinuk, cosuk, sinik, cosik,
    sinnok, cosnok, Mx, My, ux, uy, uz, vx, vy, vz ;
    int i;
    /*----------------------------------------------------------------
  update for secular gravity and atmospheric drag
----------------------------------------------------------------*/
    if (debug_flag > 1)
        printf("calling Nsgp4 with tsince= %lf\n", tsince);
    Mdf = yy->Mo + yy->Mdot * tsince;
    omgadf = yy->omegao + yy->omgdot * tsince;
    xnoddf = yy->xnodeo + yy->xnodot * tsince;
    omega = omgadf;
    Mp = Mdf;
    tsq = tsince * tsince;
    xnode = xnoddf + yy->xnodcf * tsq;
    tempa = 1. - yy->c1 * tsince;
    tempe = yy->bstar * yy->c4 * tsince;
    templ = yy->t2cof * tsq;

    if (!yy->isimp) {
        ;
        delomg = yy->omgcof * tsince;
        delm = yy->Mcof * (pow ((1. + yy->eta * cos (Mdf)), (double) 3.) - 
            yy->delMo);
        temp = delomg + delm;
        Mp = Mdf + temp;
        omega = omgadf - temp;
        tcube = tsq * tsince;
        tfour = tsince * tcube;
        tempa = tempa - yy->d2 * tsq - yy->d3 * tcube - yy->d4 * tfour;
        tempe = tempe + yy->bstar * yy->c5 * (sin (Mp) - yy->sinMo);
        templ = templ + yy->t3cof * tcube + 
            tfour * (yy->t4cof + tsince * yy->t5cof);
    }

    a = yy->aodp * tempa * tempa;
    e = yy->eo - tempe;
    L = Mp + omega + xnode + yy->xnodp * templ;
    beta = sqrt (1. - e * e);
    xn = xke / pow (a, (double) 1.5);
    /*----------------------------------------------------------------
   long period periodics 
----------------------------------------------------------------*/
    axn = e * cos (omega);
    temp = 1. / (a * beta * beta);
    Ll = temp * yy->Lcof * axn;
    aynl = temp * yy->aycof;
    Lt = L + Ll;
    ayn = e * sin (omega) + aynl;
    /*----------------------------------------------------------------
   solve keplers equation 
----------------------------------------------------------------*/

    capu = fmod2p (Lt - xnode);
    temp2 = capu;

    for (i = 0; i < 10; i++) {
        sinepw = sin (temp2);
        cosepw = cos (temp2);
        temp3 = axn * sinepw;
        temp4 = ayn * cosepw;
        temp5 = axn * cosepw;
        temp6 = ayn * sinepw;
        epw = (capu - temp4 + temp3 - temp2) / (1. - temp5 - temp6) + 
            temp2;
        if (fabs (epw - temp2) <= 1.0e-6)
            break;
        temp2 = epw;
    }

    /*----------------------------------------------------------------
  short period preliminary quantities 
----------------------------------------------------------------*/

    ecosE = temp5 + temp6;
    esinE = temp3 - temp4;
    elsq = axn * axn + ayn * ayn;
    temp = 1. - elsq;
    pl = a * temp;
    r = a * (1. - ecosE);
    temp1 = 1. / r;
    rdot = xke * sqrt (a) * esinE * temp1;
    rfdot = xke * sqrt (pl) * temp1;
    temp2 = a * temp1;
    betal = sqrt (temp);
    temp3 = 1. / (1. + betal);
    cosu = temp2 * (cosepw - axn + ayn * esinE * temp3);
    sinu = temp2 * (sinepw - ayn - axn * esinE * temp3);
    u = actan (sinu, cosu);
    sin2u = 2. * sinu * cosu;
    cos2u = 2. * cosu * cosu - 1.;
    temp = 1. / pl;
    temp1 = yy->ck2 * temp;
    temp2 = temp1 * temp;

    /*----------------------------------------------------------------
  update for short periodics 
----------------------------------------------------------------*/

    rk = r * (1. - 1.5 * temp2 * betal * yy->x3thm1) + .5 * temp1 * yy->x1mth2 * 
        cos2u;
    uk = u - .25 * temp2 * yy->x7thm1 * sin2u;
    xnodek = xnode + 1.5 * temp2 * yy->cosio * sin2u;
    xinck = yy->xincl + 1.5 * temp2 * yy->cosio * yy->sinio * cos2u;
    rdotk = rdot - xn * temp1 * yy->x1mth2 * sin2u;
    rfdotk = rfdot + xn * temp1 * (yy->x1mth2 * cos2u + 1.5 * yy->x3thm1);


    /*----------------------------------------------------------------
   orientation vectors 
----------------------------------------------------------------*/
    sinuk = sin (uk);
    cosuk = cos (uk);
    sinik = sin (xinck);
    cosik = cos (xinck);
    sinnok = sin (xnodek);
    cosnok = cos (xnodek);
    Mx = -sinnok * cosik;
    My = cosnok * cosik;
    ux = Mx * sinuk + cosnok * cosuk;
    uy = My * sinuk + sinnok * cosuk;
    uz = sinik * sinuk;
    vx = Mx * cosuk - cosnok * sinuk;
    vy = My * cosuk - sinnok * sinuk;
    vz = sinik * cosuk;

    /*----------------------------------------------------------------
   calc position and velocity, changing from earth-radii, and minutes,
   to kilometers & seconds
----------------------------------------------------------------*/
    rk *= R_EARTH;
    elem ->position[X] = rk * ux;
    elem ->position[Y] = rk * uy;
    elem ->position[Z] = rk * uz;

    elem ->velocity[X] = (rdotk * ux + rfdotk * vx) * R_EARTH / 60.;
    elem ->velocity[Y] = (rdotk * uy + rfdotk * vy) * R_EARTH / 60.;
    elem ->velocity[Z] = (rdotk * uz + rfdotk * vz) * R_EARTH / 60.;

    elem ->R = rk;
    elem ->Rdot = rdotk * R_EARTH / 60.;
    elem ->Rfdot = rfdotk * R_EARTH / 60.;
    elem ->w = fmod360 (omega * irad);
    elem ->W = fmod360 (xnodek * irad);
    elem ->u = fmod360 (uk * irad);
    elem ->i = xinck * irad;
    elem ->a = a * R_EARTH;
    elem ->e = e;
    elem ->n0 = yy->xnodp * MIN_PER_DAY / twopi;

    if (debug_flag > 1) {
        printf (" rk=%g rdotk=%g rfdotk=%g xinck=%g xnodek=%g\n",
            rk, rdotk, rfdotk, xinck * irad, xnodek * irad);
        printf (" r=%g a=%g  uk=%g L=%g Ll=%lf Lt=%g\n",
            r, a, uk * irad, L * irad, Ll * irad, Lt * irad);
        printf (" Mo=%g Mdot=%g Mdf=%g Mp=%g omega=%g xnode=%g\n",
            yy->Mo * irad, yy->Mdot * irad, Mdf * irad, Mp * irad,
            omega * irad, xnode * irad);
        for (i = 0; i < 3; i++)
            printf("r(%d) %.5lf  v(%d) %.4lf\n", i, elem->position[i],
                i, elem->velocity[i]);
    }
    if (debug_flag > 2)
        printf("leaving Nsgp4\n");
    return(0);
}
